package foraging;

import com.cyberbotics.webots.controller.DistanceSensor;
import com.cyberbotics.webots.controller.GPS;


public class Forage extends foraging_stub.Forage_stub {
    
    /* 
     * Khepera's distance sensors layout:
     * 
     *         ds2 ds3
     *      ds1       ds4
     *     ds0         ds5
     * 
     *         ds7 ds6
     */
    
    private final DistanceSensor[] ds;
    private final GPS gps;
    private final FSM fsm;
	private Detect.Detected detected;
	private float ds2, ds3;
	
	
    public Forage(DistanceSensor[] ds, GPS gps) {
        super();
        this.ds = ds;
        this.gps = gps;
        detected = null;
        fsm = new FSM(DEPART, IDLE, APPROACH, PICK_UP, DELIVER, RELEASE);
    }
    
    
    @Override
    public void init() {
        fsm.init();
    }
    
    
    @Override
    public void step() {
    	ds2 = Misc.stdRange((float)ds[2].getValue(), 15, 1000);
        ds3 = Misc.stdRange((float)ds[3].getValue(), 15, 1000);
    	Detect.Detected d = _detected.get();
	    if (d != null) {
	    	detected = d;  // consider the new detected stick
	    }
	    fsm.exec();
        if (d == null) {  // don't consider the old detected stick any more
        	detected = null;
        }
    }
    
    
    private final FSM.State IDLE = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Forage: idle");
        }
        
        @Override
        protected void exec() {
            double[] pos = gps.getValues();
            if (Misc.dist((float)pos[0], (float)pos[2], Misc.GOAL.x, Misc.GOAL.z) <= Misc.GOAL_R*1.3f) {
                // don't consider the sticks inside the goal area
                return;
            }
            if (detected != null) {  // new or previous detected stick (from the last control step)
                if (detected.front > 0) {  // approach a stick ahead
            	    transition(APPROACH);
                } else if (detected.left >= detected.right) {  // slowly turn toward a stick on the left
                	//System.out.println("Forage: slowly turn toward a stick on the left");
                	_drive.send(new Drive.Speed(-0.4f, +0.4f));
                } else if (detected.right > detected.left) {  // slowly turn toward a stick on the right
                	//System.out.println("Forage: slowly turn toward a stick on the right");
                	_drive.send(new Drive.Speed(+0.4f, -0.4f));
                }
            }
        }
    };
    
    
    private final FSM.State APPROACH = new FSM.State() {
        @Override
        protected void entry() {
        	System.out.println("Forage: approach a stick ahead");
            exec();
        }
        
        @Override
        protected void exec() {
        	if (detected == null  ||  detected.front == 0) {  // the stick has disappeared: why?
                System.out.println("Forage: the stick has disappeared!");
                transition(IDLE);
                return;
        	}
        	// detected = new or previous detected stick (from the last control step)
        	// move ahead until the front distance sensors feel the stick
        	if (ds2 + ds3 == 0) {
            	_drive.send(new Drive.Speed(1, 1));
        		return;
        	}
        	// make sure that the stick is sensed by both front distance sensors (it is almost in the middle)
        	if (ds2 == 0) {
            	//System.out.println("Forage: slowly turn toward a stick ahead on the right");
            	_drive.send(new Drive.Speed(+0.2f, -0.2f));
            	return;
        	}
        	if (ds3 == 0) {
            	//System.out.println("Forage: slowly turn toward a stick ahead on the left");
            	_drive.send(new Drive.Speed(-0.2f, +0.2f));
            	return;
        	}
        	// the stick is ahead in the middle (sensed by both front distance sensors)
            final float TARGET = 0.7f;
            float delta = TARGET - (ds2 > ds3 ? ds2 : ds3);
            // move slowly in the right position
            _drive.send(new Drive.Speed(delta/TARGET, delta/TARGET));
            if (Math.abs(delta) < 0.05) {  // the stick is in the right position -> pick it up
                transition(PICK_UP);
            }
        }
    };
    
    
    private final FSM.State PICK_UP = new FSM.State() {
        private boolean sync;
        
        @Override
        protected void entry() {
            System.out.println("Forage: pick up stick");
        	_gripperCmd.send(Gripper.CMD_PICK_UP);
            sync = false;
        }
        
        @Override
        protected void exec() {
            _drive.send(new Drive.Speed(0, 0));  // hold position
            Byte s = _gripperStatus.peek();
            if (!sync) {
                if (s == Gripper.ST_PICKING) {  // wait for the gripper to start the picking task
                    sync = true;
                }
            }
            if (sync) {
                switch (s) {
                case Gripper.ST_CARRYING:  // success
                    transition(DELIVER);
                    break;
                case Gripper.ST_IDLE:  // failure
                    transition(IDLE);
                    break;
                }
            }
        }
    };
    
    
    private final FSM.State DELIVER = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Forage: deliver stick");
            _homing.send(Misc.GOAL);
        }
        
        @Override
        protected void exec() {
            Byte s = _gripperStatus.peek();
            if (s != Gripper.ST_CARRYING) {  // the stick has fallen
                transition(IDLE);
                return;
            }
            _homing.send(Misc.GOAL);
            double[] pos = gps.getValues();
            if (Misc.dist((float)pos[0], (float)pos[2], Misc.GOAL.x, Misc.GOAL.z) <= Misc.GOAL_R*Misc.RANDOM.nextFloat()) {
                // inside the goal area -> release the stick
                transition(RELEASE);
            }
        }
    };
    
    
    private final FSM.State RELEASE = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Forage: release stick");
            _gripperCmd.send(Gripper.CMD_RELEASE);
            exec();
        }
        
        @Override
        protected void exec() {
            _drive.send(new Drive.Speed(0, 0));  // hold position
            Byte s = _gripperStatus.peek();
            if (s == Gripper.ST_IDLE) {
                transition(DEPART);
            }
        }
    };
    
    
    private final FSM.State DEPART = new FSM.State() {
        @Override
        protected void entry() {
            System.out.println("Forage: depart");
            _depart.send(Misc.GOAL);
        }
        
        @Override
        protected void exec() {
            _depart.send(Misc.GOAL);
            double[] pos = gps.getValues();
            if (Misc.dist((float)pos[0], (float)pos[2], Misc.GOAL.x, Misc.GOAL.z) > Misc.GOAL_R*2) {
                // far enough away from the goal area
                transition(IDLE);
            }
        }
    };
    
}
